module Expressions(Exp(..),exptype,freevars,tfreevars,typevars,allvars,
                   rigids,constants,freeze,thaw,flex,
                   lambda,apl,varid,
                   etaExp,etaStep,etaContract,
                   scrub,scrubtype,
                   Closure,bounds,clos,
                   evalContext,localContext,
                   unlet,nearbeta,
                   isid,freeargvars)
where

import Syntax
import Types
import Names
import List
import Utils
import Constants
import Maybe

instance Eq Exp where
   e1 == e2 
    = eq [] [] (etaContract e1) (etaContract e2)
      where eq :: [VarId] -> [VarId] -> Exp -> Exp -> Bool
            eq xs ys (Var x s) (Var y t) = s==t && 
                                           (if not (x `elem` xs || y `elem` ys) 
                                            then x==y
                                            else posin xs x == posin ys y)
            eq xs ys (Con c s) (Con d t) = s==t && c==d
            eq xs ys (Ap e1 e2 s) (Ap f1 f2 t) = s==t &&
                                                 eq xs ys e1 f1 && 
                                                 eq xs ys e2 f2
            eq xs ys (Lam p1 e1 s) (Lam p2 e2 t) = s==t &&
                                                   eq xs' ys' p1 p2 && 
                                                   eq xs' ys' e1 e2
                                                   where xs' = freevars p1 ++ xs
                                                         ys' = freevars p2 ++ ys
            eq _ _ _ _ = False
 
-- query functions on expressions

exptype :: Exp -> Type
exptype (Var _ t) = t
exptype (Con _ t) = t
exptype (Ap _ _ t) = t
exptype (Lam _ _ t) = t
exptype (Let _ _ t) = t

freevars :: Exp -> [VarId]
freevars (Var n _) = [n]
freevars (Con c _) = []
freevars (Ap e1 e2 _) = freevars e1 `union` freevars e2
freevars (Lam p e _) = freevars e \\ freevars p
freevars t = freevars (unlet t)

freeargvars :: Exp -> [VarId]
freeargvars (Var n _) = []
freeargvars (Con c _) = []
freeargvars (Lam p e _) = freeargvars e \\ freevars p
freeargvars (Ap e1 (Var n _) _) = freeargvars e1 `union` [n]
freeargvars (Ap e1 e2 _) = freeargvars e1 `union` freeargvars e2
freeargvars t = freeargvars (unlet t)

tfreevars :: Exp -> [(VarId,Type)] 
tfreevars (Var n t) = [(n,t)]
tfreevars (Con c _) = []
tfreevars (Ap e1 e2 _) = tfreevars e1 `union` tfreevars e2
tfreevars (Lam p e _) = tfreevars e \\ tfreevars p
tfreevars t = tfreevars (unlet t)

typevars :: Exp -> [TypeId]
typevars e = nub (typevars' e [])

typevars' (Var _ t) xs = vars' [] t xs
typevars' (Con _ t) xs = vars' [] t xs
typevars' (Ap e1 e2 t) xs = typevars' e1 $ typevars' e2 $ vars' [] t $ xs
typevars' (Lam e1 e2 t) xs = typevars' e1 $ typevars' e2 $ vars' [] t $ xs
typevars' t xs = typevars' (unlet t) xs

allvars :: Exp -> [VarId]
allvars (Var n _ ) = [n]
allvars (Con c _ ) = []
allvars (Ap e1 e2 _) = allvars e1 `union` allvars e2
allvars (Lam p e _) = allvars p `union` allvars e 
allvars t = allvars (unlet t)

rigids :: Exp -> [Constant]
rigids e = if null (head rs)
           then [] 
           else concat rs
           where rigs (Var y _) xs = []:xs
                 rigs (Con c _) xs = [c]:xs
                 rigs s@(Lam p e _) xs = [rigids e]
                 rigs (Ap e1 e2 _) xs = rigs e1 (rigids e2 : xs)
                 rs = rigs e []

constants :: Exp -> [Constant]
constants e = cons e []
              where cons (Var _ _) xs = xs
                    cons (Con c _) xs = c:xs
                    cons (Lam p e _) xs = cons p (cons e xs)
                    cons (Ap e1 e2 _) xs = cons e1 (cons e2 xs) 

freeze :: [VarId] -> Exp -> Exp
freeze vs (Var n t) = if n `elem` vs then Con (Constr n) t else Var n t
freeze vs (Con c t) = Con c t
freeze vs (Ap e1 e2 t) = Ap (freeze vs e1) (freeze vs e2) t
freeze vs (Lam p e t) = Lam p (freeze vs e) t

thaw :: [VarId] -> Exp -> Exp
thaw vs (Var c t) = Var c t
thaw vs (Con (Constr n) t) = if n `elem` vs then Var n t else Con (Constr n) t
thaw vs (Con c t) = Con c t
thaw vs (Ap e1 e2 t) = Ap (thaw vs e1) (thaw vs e2) t
thaw vs (Lam p e t) = Lam p (thaw vs e) t

flex :: Exp -> Bool
flex e = f [] e
         where f xs (Var x _) = not (x `elem` xs)
               f xs (Con _ _) = False
               f xs (Ap e1 _ _) = f xs e1
               f xs (Lam p e _) = True

-- composing typed expressions

lambda (x,t) e = Lam (varid (x,t)) e (arrow t (exptype e)) 

apl e1 e2 = Ap e1 e2 (restype (exptype e1))

varid (x,t) = Var x t


-- eta expansion

etaExp :: Exp -> Name Exp
etaExp (Lam x p t) = recexpand (Lam x p t) 
etaExp e = case (exptype e) of
             t@(TCon "Arrow" [t1,t2]) -> do x <- freshVar
                                            e' <- etaExp (Lam (Var x t1) (Ap e (Var x t1) t2) t)
                                            return e'
             _ -> recexpand e 

recexpand e@(Var _ _) = return e
recexpand e@(Con _ _) = return e
recexpand (Ap e1 e2 t) = do e1' <- recexpand e1
                            e2' <- etaExp e2
                            return (Ap e1' e2' t)
recexpand (Lam p e t) = do e' <- etaExp e
                           return (Lam p e' t)


etaStep :: Exp -> Name Exp
etaStep e = case (exptype e) of
             t@(TCon "Arrow" [t1,t2]) 
                      -> do x <- freshVar
                            return (Lam (Var x t1) (Ap e (Var x t1) t2) t)
             _        -> return e 

etaContract :: Exp -> Exp
etaContract (Lam p (Ap e q u) v) 
       = if p==q && null (intersect (freevars p) (freevars e))
         then e'
         else Lam p (Ap e' q' u) v
         where e' = etaContract e
               q' = etaContract q
etaContract (Lam p e t) = case e' of
                            (Ap _ _ _) -> etaContract (Lam p e' t)
                            _ -> Lam p e' t
                          where e' = etaContract e
etaContract (Ap e1 e2 t) = Ap (etaContract e1) (etaContract e2) t 
etaContract e = e

unlet (Let (DList (OneDef (Fun p e) ds _)) b t) = 
        Ap (Lam p b' (arrow (exptype p) (exptype b'))) (unlet e) t
                      where b' = unlet (Let (DList ds) b t)
unlet (Let (DList NoDef) e t) = unlet e
unlet (Var x t) = Var x t 
unlet (Con c t) = Con c t
unlet (Ap e1 e2 t) = Ap (unlet e1) (unlet e2) t 
unlet (Lam e1 e2 t) = Lam e1 (unlet e2) t 
                       
					
nearbeta (Let (DList ds) b t) = lets (nearbetas ds) (nearbeta b) t
nearbeta (Var x t) = Var x t 
nearbeta (Con c t) = Con c t
nearbeta (Ap (Lam p e1 _) e2 t) = lets (OneDef (Fun p (nearbeta e2)) NoDef 0) (nearbeta e1) t
nearbeta (Ap e1 e2 t) = Ap (nearbeta e1) (nearbeta e2) t
nearbeta (Lam e1 e2 t) = Lam e1 (nearbeta e2) t

nearbetas NoDef = NoDef
nearbetas (OneDef (Fun p e) ds n) = OneDef (Fun p (nearbeta e)) (nearbetas ds) n

catDefs NoDef ds = ds
catDefs (OneDef d ds n) es = OneDef d (catDefs ds es) n

lets ds (Let (DList es) b _) t = Let (DList (catDefs ds es)) b t 
lets ds e t = Let (DList ds) e t
                       

-- hygienic naming

scrub :: Exp -> Rename Exp
scrub (Var n t)  = do n' <- rename n
                      t' <- scrubtype t
                      return (Var n' t')
scrub (Con c t) = do t' <- scrubtype t
                     return (Con c t')
scrub (Lam x e t) = local (do xs <- mapM newname (freevars x)
                              x' <- scrub x
                              e' <- scrub e 
                              t' <- scrubtype t
                              return (Lam x' e' t'))
scrub (Ap e1 e2 t) = do e1' <- scrub e1 
                        e2' <- scrub e2
                        t'  <- scrubtype t
                        return (Ap e1' e2' t')


scrubtype :: Type -> Rename Type
scrubtype (Tid x) = do x' <- retype x
                       return (Tid x')
scrubtype (TCon x ts) = do ts' <- mapM scrubtype ts
                           return (TCon x ts')
scrubtype (TAbs xs t) = do xs' <- mapM retype xs
                           t' <- scrubtype t
                           return (TAbs xs' t')

type Closure = (Exp,[(VarId,Type)])
bounds (e,xs) = xs
clos e = (e,[])


evalContext :: Closure -> Name a -> a
evalContext cl x = evalName (supplyContext cl supply) x

localContext :: Closure -> Name a -> Name a
localContext cl x = localName (supplyContext cl) x

supplyContext :: Closure -> NameSupply -> NameSupply
supplyContext (e,bs) = suppex (map fst bs++allvars e) 
                           (typevars e ++ concat (map (vars.snd) bs))

-- identity checking

isid (Lam (Var x _) (Var y _) _) = x==y
isid _ = False

