module Inversion where

import Monad
import Pan

{-

Gradually invert, with d as fixpoint as z as progress indicator.
Satisfies for some e,

  ginv r 0    z = z,   for all r,z   (start: identity)
  ginv r 1    z = e/z, for all r,z   (end: inversion)
  ginv r frac r = r,   for all frac  (fixpoint)

Solution: Second and third conditions imply r = e/r, so e = r^2.  So use
ginv r frac z = r * (z/r)^(1-2*frac)

But I really want the *inverse* of (ginv d z), because we use our
transforms backwards.

If z' = r * (z/r)^(1-2*frac), then

   z'/r = (z/r)^(1-2*frac)

   (z'/r)^(1/(1-2*frac)) = (z/r)

   r * (z'/r)^(1/(1-2*frac)) = z

-}

ginv :: FloatE -> FloatE -> (FloatE -> FloatE)
ginv r frac z = r * (z/r)**(1/(1-2*frac))

gradualInvert :: FloatE -> FractionE -> Filter a
gradualInvert r frac = imXf $ radialXf (ginv r frac)

invertFUI :: UI (FractionE -> Filter a)
invertFUI = liftM gradualInvert $
            slider "fixpoint radius" (10,150) 60

invertUI :: UI (Filter a)
invertUI = atMouse $ invertFUI `ap` progressFromUI 0.25

bigChecker = uscale 3 cb1

powSlider :: Label -> Double -> UI FloatE
powSlider label v0 = liftM (2**) $ slider label (-15,30) v0
     

invertCheckerUI :: UI ImageC
invertCheckerUI = liftM ($ bigChecker) invertUI

{-
-- No longer do the cropping
invertCheckerUI = do imF      <- invertUI
                     cropSize <- powSlider "outer crop power" 15
                     return $
                       imF (crop (uscale cropSize ubox) bigChecker)
 where
  ubox (x,y) = abs x <* 1/2 &&* abs y <* 1/2
-}   

moveInvChecker v = gradualInvert 50 1 (translate v cb1)

-- mousyInvCheckerUI = do
--   sens <- liftM (2**) $ slider "mouse sensitivity power" (-5,2) (-3)
--   liftM (moveInvChecker . uscaleP sens) mouseInput


--------------------------------------------------

makeAll = makeAllFor allBackends

makeAllFor bends = do
  compileImageFor  bends "inverting checker"      invertCheckerUI
  compileImageFor  bends "inverted b/w checker"   (return $ gradualInvert 50 1 cb0)
  compileImageBFor bends "inverting moving checker linear"
             (return (moveInvChecker . \ t -> (3*t,3*t)))
  compileImageBFor bends "inverting moving checker circular"
             (return (moveInvChecker . \ t -> fromPolar (50,t)))
  --compileImageFor  bends "mousy inverted checker" mousyInvCheckerUI
  compileFilterFor bends "inversion"              invertUI

