/*
 *  The following code implements a value-based wrapper around
 *  the controls and containers provided by UILib (uiLib.cpp) 
 *
 *  The UI is described by way of a hierarchical data structure,
 *  which is then instantiated to produce an actual GUI. (The
 *  reason for using a separate data structure is that it
 *  simplifies and hides away details associated with the
 *  creation of UI elements.)
 *
 *  The UI notifies the underlying 'app' of changes by overwriting
 *  variables with the updated value, followed by a call to an
 *  invalidation/refresh routine, which causes the view(s) to be
 *  updated.
 * 
 *  -- sof 2/00
 */
#include <stdarg.h>
#include <stdio.h>
#include "stdafx.h"

#if 0
#include "ImageViewer.h"
#include "PhSImage.h"
#include "uiLib.h"
#include "Sizer.h"
#endif
#include "uiSpec.h"


// 
// FUNCTION: mkVBox
//
// PURPOSE: create a vertical layout container, along with
//          a label describing the grouping.
//
// COMMENTS: Use C's varargs here let us specify the children of the
//           box 'inline' without having to stuff them into an array first. 
//
UI*
mkVBox(char* lab, int no, ...)
{
    va_list ap;
    int i;
    UI* ptr = (UI*)malloc(sizeof(UI));
    VBoxInfo* cb = (VBoxInfo*)malloc(sizeof(VBoxInfo) + (max(0,(no-1)))*sizeof(UI*));

    cb->label = lab;
    cb->noChildren = no;

    // process the children. More than 'no' children might be passed, but
    // we will only be adding 'no' of them here.
    va_start(ap, no);

    for (i=0; i < no ; i++) {
      cb->children[i] = (UI*)va_arg(ap, UI*);
    }

    va_end(ap);

    // fill in kind and element-specific data.
    ptr->kind = VBox;
    ptr->info.vbox = cb;

    return ptr;
}

// 
// FUNCTION: mkHBox
//
// PURPOSE: create a horizontal layout container, along with
//          a label describing the grouping.
//
UI*
mkHBox(char* lab, int no, ...)
{
    va_list ap;
    int i;
    UI* ptr = (UI*)malloc(sizeof(UI));
    VBoxInfo* cb = (VBoxInfo*)malloc(sizeof(HBoxInfo) + max(0,no-1)*sizeof(UI*));

    cb->label = lab;
    cb->noChildren = no;

    // Adding the children - see mkVBox remark.
    va_start(ap, no);

    for (i=0; i < no ; i++) {
      cb->children[i] = (UI*)va_arg(ap, UI*);
    }

    va_end(ap);

    ptr->kind = HBox;
    ptr->info.vbox = cb;

    return ptr;
}

// 
// FUNCTION: mkSlider
//
// PURPOSE: create a slider, i.e., a trackbar and an edit box
//          displaying current value. The slider is bounded,
//          so the (float valued) extreme values also need to be
//          supplied.
//
UI*
mkSlider(char* lab,
	 float min,
	 float max,
	 float initVal,
	 FloatUpdater* pUpd,
	 void* pParam)
{
    UI* ptr = (UI*)malloc(sizeof(UI));
    SliderInfo* si = (SliderInfo*)malloc(sizeof(SliderInfo));

    si->label = lab;
    si->max   = max;
    si->min   = min;
    si->initVal = initVal;
    si->pUpdateValue = pUpd;
    si->pParam = pParam;

    ptr->kind = Slider;
    ptr->info.slider = si;
	
    return ptr;
}

// 
// FUNCTION: mkISlider
//
// PURPOSE: create an integer slider, i.e., a trackbar and an edit box
//          displaying current value. The slider is bounded,
//          so the (int valued) extreme values also need to be
//          supplied.
//
UI*
mkISlider(char* lab,
	 int min,
	 int max,
	 int initVal,
	 IntUpdater* pUpd,
	 void* pParam)

{
    UI* ptr = (UI*)malloc(sizeof(UI));
    ISliderInfo* si = (ISliderInfo*)malloc(sizeof(ISliderInfo));

    si->label = lab;
    si->max   = max;
    si->min   = min;
    si->initVal = initVal;
    si->pUpdateValue = pUpd;
    si->pParam = pParam;

    ptr->kind = ISlider;
    ptr->info.islider = si;
	
    return ptr;
}

// 
// FUNCTION: mkTick
//
// PURPOSE: create a 2-state check/tick box.
//
UI*
mkTick(char* lab,
       int defVal,
       IntUpdater* pUpd,
       void* pParam)
{
    UI* ptr = (UI*)malloc(sizeof(UI));
    TickInfo* ti = (TickInfo*)malloc(sizeof(TickInfo));

    ti->label        = lab;
    ti->defValue     = defVal;
    ti->pUpdateValue = pUpd;
    ti->pParam       = pParam;

    ptr->kind = Tick;
    ptr->info.tick  = ti;

    return ptr;
}

// 
// FUNCTION: mkMouseInput
//
// PURPOSE: create a mouse input 'ui' element
//
UI*
mkMouseInput(FloatUpdater* pUpd, void* pParamX, void* pParamY)
{
    UI* ptr = (UI*)malloc(sizeof(UI));
    MouseInputInfo* mi = (MouseInputInfo*)malloc(sizeof(MouseInputInfo));

    mi->updMouse = pUpd;
    mi->pParamX  = pParamX;
    mi->pParamY  = pParamY;

    ptr->kind = MouseInput;
    ptr->info.mouseInput = mi;

    return ptr;
}

// 
// FUNCTION: mkButton
//
// PURPOSE: create a push button.
//
UI*
mkButton(char* lab, int val)
{
    UI* ptr = (UI*)malloc(sizeof(UI));
    ButtonInfo* bi = (ButtonInfo*)malloc(sizeof(ButtonInfo));

    bi->label = lab;
    bi->value  = val;

    ptr->kind = Button;
    ptr->info.button = bi;

    return ptr;
}

// 
// FUNCTION: mkLabel
//
// PURPOSE: create a static label
//
UI*
mkLabel(char* lab)
{
    UI* ptr = (UI*)malloc(sizeof(UI));
    LabelInfo* li = (LabelInfo*)malloc(sizeof(LabelInfo));

    li->label = lab;

    ptr->kind = Label;
    ptr->info.label = li;

    return ptr;
}

// 
// FUNCTION: mkResizeable
//
// PURPOSE: changes the resize properties of a UI elem.
//
UI*
mkResizeable(BOOL hResiz, BOOL vResiz, UI* elem)
{
    UI* ptr = (UI*)malloc(sizeof(UI));
    ResizeInfo* ri = (ResizeInfo*)malloc(sizeof(ResizeInfo));

    ri->hResize = hResiz;
    ri->vResize = vResiz;
    ri->uiElem  = elem;

    ptr->kind = Resizeable;
    ptr->info.resize = ri;

    return ptr;
}

// 
// FUNCTION: mkImportImage
//
// PURPOSE: create a control that lets the user select
//          an image to import, at run-time. The custom
//          control consist of an edit field along with 
//          a button that fires up an OpenFile common dialog.
//
UI*
mkImportImage(char* lab,
	      char* defImg,
	      IntUpdater* pUpd,
	      void* pParamBits,
	      void* pParamWidth,
	      void* pParamHeight)
{
    UI* ptr = (UI*)malloc(sizeof(UI));
    ImageInfo* ii = (ImageInfo*)malloc(sizeof(ImageInfo));

    ii->label        = lab;
    ii->defImg       = defImg;
    ii->pUpdateValue = pUpd;
    ii->pParamBits   = pParamBits;
    ii->pParamWidth  = pParamWidth;
    ii->pParamHeight = pParamHeight;
    
    ptr->kind = ImportImage;
    ptr->info.image  = ii;

    return ptr;
}

// 
// FUNCTION: mkRadio
//
// PURPOSE: create a radio box, providing exclusive choice among
//          'no' different entries. Selecting a label causes the
//          underlying app to be notified. The value reported is
//          the position of the label in the list of arguments
//          being passed to 'mkRadio', i.e., in the range [0..(no-1)]
//
UI*
mkRadio(char* lab,
	int defVal,
	IntUpdater* pUpd,
	void* pParam,
	int no,
	...)
{
    va_list ap;
    int i;
    UI* ptr = (UI*)malloc(sizeof(UI));
    RadioInfo* cb = (RadioInfo*)malloc(sizeof(RadioInfo) + max(0,no-1)*sizeof(char*));

    cb->label        = lab;
    cb->defValue     = defVal;
    cb->pUpdateValue = pUpd;
    cb->pParam       = pParam;
    cb->noItems      = no;

    va_start(ap, no);

    for (i=0; i < no ; i++) {
      cb->labels[i] = (char*)va_arg(ap, char*);
    }

    va_end(ap);
    
    ptr->kind = Radio;
    ptr->info.radio = cb;

    return ptr;
}

// 
// FUNCTION: mkViewWindow
//
// PURPOSE: create a window that will render a DisplayProc.
//
// COMMENTS: The pointer to an HWND passed in is used to set up the 
//           connection between the view's invalidation routine
//           and the view window.
// 
UI*
mkViewWindow(PanImageDisplay* disp,
	     BOOL tVarying,
	     int nWidth,
	     int nHeight)
{
    UI* ptr = (UI*)malloc(sizeof(UI));
    DisplayInfo* di = (DisplayInfo*)malloc(sizeof(DisplayInfo));

    di->displayer = disp;
    di->timeVarying = tVarying;
    di->nWidth = nWidth;
    di->nHeight = nHeight;

    ptr->kind = Display;
    ptr->info.display = di;

    return ptr;
}

// 
// FUNCTION: mkPhSViewWindow
//
// PURPOSE: create a window that will render a PhSDisplayer.
//
// 
UI*
mkPhSViewWindow (void* pImg, int nWidth, int nHeight)
{
    UI* ptr = (UI*)malloc(sizeof(UI));
    PhSDisplayInfo* pi = (PhSDisplayInfo*)malloc(sizeof(PhSDisplayInfo));

    pi->pImg = pImg;
    pi->nWidth = nWidth;
    pi->nHeight = nHeight;

    ptr->kind = PhSDisplay;
    ptr->info.phsdisplay = pi;

    return ptr;
}


// 
// FUNCTION: mkPhSViewWindow
//
// PURPOSE: create a control that lets user type in a string
//          and render it using a user selected font.
//
UI*
mkTextBox
   ( char*       lab,
     char*       defVal,
     IntUpdater* pUpd,
     void*       pParamBits,
     void*       pParamWidth,
     void*       pParamHeight
   )
{
    UI* ptr = (UI*)malloc(sizeof(UI));
    TextInfo* ti = (TextInfo*)malloc(sizeof(TextInfo));

    ti->label        = lab;
    ti->defVal       = defVal;
    ti->pUpdateValue = pUpd;
    ti->pParamBits   = pParamBits;
    ti->pParamWidth  = pParamWidth;
    ti->pParamHeight = pParamHeight;

    ptr->kind = TextBox;
    ptr->info.textBoxInfo = ti;

    return ptr;

}


extern HINSTANCE hInst;

#if 0
extern int         handleMouseInput;
extern float*      g_pX;
extern float*      g_pY;
extern InvalidProc g_iProc;

//
// A change in any of the UI settings requires that
// the view(s) are updated. Do this via InvalidateViewer(),
// which will invalidate all the view windows & cause a 
// repaint.
//
// We barely have a story for more than one view window, so
// limiting InvalidateViewer() to only work over MAX_VIEWERS,
// shouldn't constrain us in any practical way.
// 
#define MAX_VIEWERS 16

static HWND        viewHWNDS[MAX_VIEWERS];
static int         viewIdx = 0;
static PhSImage*   pImage = NULL;

static void InvalidateViewer()
{
  int i;
  
  if (pImage) {
    pImage->Update();
  }
  
  for(i=0;i< viewIdx ; i++) {
    InvalidateRect(viewHWNDS[i], NULL, FALSE);
  }
}

//
// FUNCTION: HWND createUI(UI*)
//
// PURPOSE: Given a tree describing the UI to be created, instantiate
//          it.
// 
// COMMENTS: Distinguishing between the description of a UI tree and
//           its realisation lets us hide away irrelevant detail and
//           the downward propagation of information (i.e., passing
//           a parent's HWND to its children.)
// 
HWND
createUI(HWND hWndParent, UI* pSpec)
{

    switch (pSpec->kind) {


    case HBox:
      { 
	HWND hWndBox;
	int i;

	// Creating the box - the 'magic' numbers specifying the size
	// and position will be updated once all the children have been
	// added and the size of the box has been recalculated.
	hWndBox =
	  CreateHBox ( pSpec->info.hbox->label
                     , 0
                     , 10
                     , 10
                     , 50
                     , 50
                     , hWndParent
		     , (HMENU)-1
		     , hInst
		     , TRUE
		     );

	for (i=0;i < pSpec->info.hbox->noChildren; i++) {
	  createUI(hWndBox, pSpec->info.hbox->children[i]);
	}
	RecalcSize(hWndBox);
	return hWndBox;

      }
    case VBox:
      { 
	HWND hWndBox;
	int i;

	hWndBox =
	  CreateVBox ( pSpec->info.hbox->label
		     , 0
		     , 10
		     , 10
		     , 50
		     , 50
		     , hWndParent
		     , (HMENU)-1
		     , hInst
		     , TRUE
		       );

	for (i=0;i < pSpec->info.hbox->noChildren; i++) {
	  createUI(hWndBox, pSpec->info.hbox->children[i]);
	}
	RecalcSize(hWndBox);
	return hWndBox;
      }
    case Tick:
      return createTick ( hWndParent
			, pSpec->info.tick->label
			, 50, 20
			, pSpec->info.tick->defValue
			, pSpec->info.tick->pUpdateValue
			, pSpec->info.tick->pParam
			, InvalidateViewer
			);

    case MouseInput: {
         handleMouseInput = 1;
	 g_pX = pSpec->info.mouseInput->pX;
	 g_pY = pSpec->info.mouseInput->pY;
	 g_iProc = InvalidateViewer;
	 return NULL;
    }

    case Button:
      return createButton ( hWndParent
			, pSpec->info.button->label
			, pSpec->info.button->value
			, 50, 30
			);

    case Label:
      return createLabel ( hWndParent
			 , pSpec->info.button->label
			 , 50, 30 // undesirable to have to give the dims here. ToDo: better.
			 );

    case Resizeable:
      return createResizeable ( pSpec->info.resize->hResize
			      , pSpec->info.resize->vResize
			      , createUI (hWndParent, pSpec->info.resize->uiElem)
			      );

    case Radio:
      return createRadioBox ( hWndParent
			    , pSpec->info.radio->label
			    , (LPCTSTR*)pSpec->info.radio->labels
			    , pSpec->info.radio->noItems
			    , pSpec->info.radio->defValue
 		            , pSpec->info.radio->pUpdateValue
 		            , pSpec->info.radio->pParam
			    , InvalidateViewer
			    );
    case Slider:
      return createSlider ( hWndParent
			  , pSpec->info.slider->label
			  , pSpec->info.slider->min
			  , pSpec->info.slider->max
			  , pSpec->info.slider->initVal
			  , pSpec->info.slider->pUpdateValue
			  , pSpec->info.slider->pParam
			  , InvalidateViewer
			  );

    case ISlider:
      return createISlider ( hWndParent
			   , pSpec->info.islider->label
			   , pSpec->info.islider->min
			   , pSpec->info.islider->max
			   , pSpec->info.islider->initVal
			   , pSpec->info.islider->pUpdateValue
			   , pSpec->info.islider->pParam
			   , InvalidateViewer
			   );

    case ImportImage:
      return createIImage ( hWndParent
			  , pSpec->info.image->label
			  , pSpec->info.image->defImg
			  , 70
			  , 20
			  , pSpec->info.image->pUpdateValue
			  , pSpec->info.image->pParamBits
			  , pSpec->info.image->pParamWidth
			  , pSpec->info.image->pParamHeight
			  , InvalidateViewer
			  );
    case TextBox:
      return createTextBox 
                ( hWndParent
                , pSpec->info.textBoxInfo->label
                , pSpec->info.textBoxInfo->defVal
                , 70
                , 20
                , pSpec->info.textBoxInfo->pUpdateValue
                , pSpec->info.textBoxInfo->pParamBits
                , pSpec->info.textBoxInfo->pParamWidth
                , pSpec->info.textBoxInfo->pParamHeight
                , InvalidateViewer
                );

    case Display:
      {
	HWND hWnd;
	
	hWnd = createViewer 
	                  ( hWndParent
			  , pSpec->info.display->displayer
			  , pSpec->info.display->timeVarying
			  , pSpec->info.display->nWidth
			  , pSpec->info.display->nHeight
			  );

        if (viewIdx < MAX_VIEWERS) {
	  viewHWNDS[viewIdx] = hWnd;
	  viewIdx++;
	  return hWnd;
	} else {
	  MessageBox(NULL,
		     TEXT("Ran out of viewer windows; some viewer windows won't be auto-redisplayed"),
		     TEXT("CreateUI"), MB_OK);
	  return hWnd;
	}
      }     

    case PhSDisplay:
      {
	HWND hWnd;
	
	pImage = (PhSImage*)pSpec->info.phsdisplay->pImg;
	
        hWnd = createPhSViewer
                          ( hWndParent
			  , (PhSImage*)pSpec->info.phsdisplay->pImg
			  , pSpec->info.phsdisplay->nWidth
			  , pSpec->info.phsdisplay->nHeight
			  );

        if (viewIdx < MAX_VIEWERS) {
	  viewHWNDS[viewIdx] = hWnd;
	  viewIdx++;
	  return hWnd;
	} else {
	  MessageBox(NULL,
		     TEXT("Ran out of viewer windows; some viewer windows won't be auto-redisplayed"),
		     TEXT("CreateUI"), MB_OK);
	  return hWnd;
	}
      }
    }

    return (HWND)NULL;
}
#endif

void freeUI(UI* pUI)
{
    // Only the UI and the *Info structs are dynamically
    // allocated - i.e., the char* are assumed to be static.
    // (same goes for the object pointers - they're assumed
    // to be free using other mechs.)
    switch (pUI->kind) {
    case HBox:
        {
            int i, no = pUI->info.hbox->noChildren;

            for (i=0; i < no ; i++) {
                freeUI(pUI->info.hbox->children[i]);
            }
            free(pUI->info.hbox);
            free(pUI);
            return;
        }
    case VBox:
        {
            int i, no = pUI->info.vbox->noChildren;

            for (i=0; i < no ; i++) {
                freeUI(pUI->info.vbox->children[i]);
            }
            free(pUI->info.vbox);
            free(pUI);
            return;
        }
    case Slider:
        {
            free(pUI->info.slider);
            free(pUI);
            return;
        }
    case ISlider:
        {
            free(pUI->info.islider);
            free(pUI);
            return;
        }
    case Tick:
        {
            free(pUI->info.tick);
            free(pUI);
            return;
        }
    case Button:
        {
            free(pUI->info.button);
            free(pUI);
            return;
        }
    case Label:
        {
            free(pUI->info.label);
            free(pUI);
            return;
        }
    case MouseInput:
        {
            free(pUI->info.mouseInput);
            free(pUI);
	    return;
        }
    case Radio:
        {
	  free(pUI->info.radio);
	  free(pUI);
	  return;
        }
    case TextBox:
        {
            free(pUI->info.textBoxInfo);
            free(pUI);
            return;
        }
    case ImportImage:
        {
            free(pUI->info.image);
            free(pUI);
            return;
        }
    case Resizeable:
        {   
            free(pUI->info.resize);
            free(pUI);
            return;
        }
    case Display:
        {
            free(pUI->info.display);
            free(pUI);
            return;
        }
    case PhSDisplay:
        {
            free(pUI->info.phsdisplay);
            free(pUI);
            return;
        }
    default:
        {
#if 1 //DEBUG
            TCHAR msg[255];
            _stprintf(msg, TEXT("unhandled type %d"), pUI->kind);
            MessageBox(NULL, msg, TEXT("freeUI"), MB_OK);
#endif
            return;
        }
    }
}
