-- Support for loading bitmap files.  Adapted from Alastair Reid's
-- GraphicsBitmap.hs.

module LoadBitmap 
	  (loadBMPHeader
	  , loadBMPBitsPal
	  , getPicDimensions
	  ) where

-- the BMP mungers are now implemented in externally (in utils.c).
import  -- trick 'ghc -M'
	PanUtils

{- Win32 library based solution:

-- To do: replace this monolithic import with more selective ones.
import  -- fool mkdependHS
       qualified Win32

import Word
import Addr( plusAddr )
import Int
import IO (bracket)

-- Work around temporary Hugs/GHC inconsistency
{- BEGIN_GHC_ONLY
readF = Win32.win32_ReadFile
   END_GHC_ONLY -}
{- BEGIN_NOT_FOR_GHC -}
readF = Win32.readFile
{- END_NOT_FOR_GHC= -}


----------------------------------------------------------------
-- Reading bitmaps from files
----------------------------------------------------------------

-- ToDo: the "bits" read are never freed but I think we can free them
-- as soon as we call createDIBitmap.

-- Summary of the Win32 documentation on BMP files:
--
-- A bitmap file consists of:
-- 
-- 	   +-------------------+
-- 	   | BITMAPFILEHEADER  |
-- 	   +-------------------+
-- 	   | BITMAPINFOHEADER  |
-- 	   +-------------------+
-- 	   | Rgbquad array     |
-- 	   +-------------------+
-- 	   | Color-index array |
-- 	   +-------------------+
-- 
-- The file header tells you the size of the file and the offset of the
-- bitmap data from the header start.
-- 
-- The info header specifies the width and height, the colour format,
-- compression mode, number of bytes of data, resolution and the number
-- of colours.
-- 
-- The RGBQUAD array is a palette.
-- 
-- The Color-index array is the actual bitmap.

loadBMPHeader :: String -> IO Win32.BITMAPINFOHEADER

loadBMPHeader fileName = 
  bracket 
    (Win32.createFile fileName 
	  	      Win32.gENERIC_READ Win32.fILE_SHARE_READ Nothing 
		      Win32.oPEN_EXISTING Win32.fILE_ATTRIBUTE_NORMAL Nothing)
    Win32.closeHandle
    $ \ file -> do
  (offset, size) <- readFileHeader file
  pHeader        <- readBMPHeader file offset size
  Win32.getBITMAPINFOHEADER_ pHeader


loadBMPBits :: String -> IO Win32.LPVOID

loadBMPBits fileName = 
  bracket 
    (Win32.createFile fileName 
	  	      Win32.gENERIC_READ Win32.fILE_SHARE_READ Nothing 
		      Win32.oPEN_EXISTING Win32.fILE_ATTRIBUTE_NORMAL Nothing)
    Win32.closeHandle
    $ \ file -> do
  (offset, size) <- readFileHeader file
  readBMPBits file offset size

loadBMPPalette :: Word32 -> String -> IO Addr
loadBMPPalette clrUsed fileName = 
  bracket 
    (Win32.createFile fileName 
	  	      Win32.gENERIC_READ Win32.fILE_SHARE_READ Nothing 
		      Win32.oPEN_EXISTING Win32.fILE_ATTRIBUTE_NORMAL Nothing)
    Win32.closeHandle
    $ \ file -> do

    -- get at the byte offset from the end of the FILEHEADER to the
    -- start of the bitmap bits..
  (offset, size) <- readFileHeader file
  fileHeader <- Win32.malloc offset
  read <- readF file fileHeader offset Nothing
  assert (read == offset) "Bitmap file ended unexpectedly"
    -- ...and adjust the returned pointer to point past the 
    -- size of the INFOHEADER (whatever it is) & have it point
    -- to the start of the palette.
    --
    -- ToDo: WHY do I need to adjust the offset down by 1 byte here?!?
  return (fileHeader `plusAddr` toInt (offset - 4 * clrUsed - 1))

readFileHeader :: Win32.HANDLE -> IO (Word32, Word32)
readFileHeader file = do
  -- read the file header
  fileHeader <- Win32.malloc Win32.sizeofLPBITMAPFILEHEADER
  read <- readF file fileHeader
            Win32.sizeofLPBITMAPFILEHEADER Nothing
  assert (read == Win32.sizeofLPBITMAPFILEHEADER) "Bitmap file lacks header"
  -- check the tag and get the size
  (tag, size, r1, r2, offset) <- Win32.getBITMAPFILEHEADER fileHeader
  assert (tag == intToWord16 (fromEnum 'B' + 256 * fromEnum 'M')) 
         "Bitmap file lacks tag"
  assert (r1 == 0 && r2 == 0) 
         "Bitmap header contains non-zero reserved words"
  Win32.free fileHeader
  return ( offset - Win32.sizeofLPBITMAPFILEHEADER
         , size - Win32.sizeofLPBITMAPFILEHEADER
         )

readBMPHeader :: Win32.HANDLE -> Word32 -> Word32 -> IO Win32.LPBITMAPINFOHEADER
readBMPHeader file offset size = do
  header <- Win32.malloc offset
  read <- readF file header offset Nothing
  assert (read < size) "Bitmap file ended unexpectedly"
  return header


-- read the bits out of the rest of the file
-- assumes that you've just read the file header
readBMPBits :: Win32.HANDLE -> Word32 -> Word32 -> IO Win32.LPVOID
readBMPBits file offset size = do
  header <- Win32.malloc size
  read <- readF file header size Nothing
  assert (read == size) "Bitmap file ended unexpectedly"
  -- Skip past the header, returning just the bits.  Would be less
  -- wasteful to skip over the header.
  return (header `plusAddr` word32ToInt offset)


-- In the development system, this might print the error message
-- if the assertion fails.
{-
assert :: Bool -> String -> IO ()
assert _ _ = return ()
-}

assert True _ = return ()
assert False why = ioError (userError ("Assertion failed:" ++ why))

testReadBMPFile :: String -> IO ()
testReadBMPFile fileName = do
  (_, w, h, planes, bitCount, compress, imSize, ppmx, ppmy, clrUsed, clrImp) <-
    loadBMPHeader fileName
  print ((w,h,planes,bitCount,compress),
         (imSize, ppmx, ppmy,clrUsed))
  bits <- loadBMPBits fileName
  print bits

t1 = testReadBMPFile "../import/fraidy-cat.bmp"
-}
